/*!
 * \file Webserver Source File *
 *
 *  Webserver hardware is a MicroChip ENC28J60.
 *  datasheet:      ethernet datasheet
 *
 *  Helpful site for describing multidemensional arrays:  ac:Multi_Array_Site
 *
 */

#include "webserver.h"
//#include "rtctime.h"
#include "built_in.h"
#include "__Lib_FAT32.h"

/*****************************
     Globals / All globals have static duration
*****************************/
// mE ethernet NIC pinout
sfr sbit Net_Ethernet_28j60_Rst at PORTA0_bit;
sfr sbit Net_Ethernet_28j60_CS  at PORTC2_bit;
sfr sbit Net_Ethernet_28j60_Rst_Direction at DDA0_bit;
sfr sbit Net_Ethernet_28j60_CS_Direction  at DDC2_bit;
// SD Card port / pins
sbit Mmc_Chip_Select            at PORTC4_bit;
sbit Mmc_Chip_Select_Direction  at DDC4_bit;
sbit Mmc_Insert_Detect          at PORTC5_bit;
sbit Mmc_Insert_Direction       at DDC5_bit;
// Ethernet variables
unsigned char myMacAddr[6] = {0x00, 0x14, 0xA5, 0x76, 0x19, 0x3f};  // my MAC address
unsigned char myIpAddr[4]  = {192, 168,  1, 150};                   // my IP address
unsigned char gwIpAddr[4]  = {192, 168,  1,  1};                    // gateway (router) IP address
unsigned char ipMask[4]    = {255, 255, 255,  0};                   // network mask (for example : 255.255.255.0)
unsigned char dnsIpAddr[4] = {192, 168,  20,  1};                   // DNS server IP address
unsigned short portNumber = 80;
int dhcp_mode = DHCP_OFF;
/*!
 * \struct webSocket_desc
 * \brief Websocket struct with file buffer information from SD card
 *
 * Multiple webSocket_desc can be used to speed up transfers, however
 * the memory requirement for this can grow quickly based on the buffer sizes.
 */
struct webSocket_desc{
     SOCKET_28j60_Dsc *socket;     /*!< Socket for TCP connection */
     char mark;                    /*!< Mark if socket is in use or free */
     unsigned long position;       /*!< Position of where in the packet it is has transmitted */
     char fileName[COMMANDSIZE];   /*!< The file name associated with the socket */
     __HANDLE fileHandle;
     int fileType;
     unsigned long fileSize;       /*!< The file size in bytes */
     char buffer[DATABUFFERSIZE];  /*!< Buffer for reading from the file and ready to send */
     unsigned int bufPos;          /*!< Where in the buffer has it sent, when at the end, time for recharge if not EOF */
};

typedef struct webSocket_desc webSocket_desc_t;
// Misc globals
webSocket_desc_t webSoc[MAX_SOCKETS];                                         // Websockets for Webserver
volatile bit status;
volatile bit sdCardInsert;
// UDP Packet responses
struct SNTP_Response{
     long epoch;
     unsigned char serverFlags;
     unsigned char serverStratum;
     char serverPrecision;
};
// Function pointer struct
struct webServer_desc Webserver = {
       &webserverInit,
       &SDCardInit,
       &doPacket,
       &logEvent,
       &setIP,
       &setSubIP,
       &setGW,
       &setDNS,
       &setDHCP,
       &setPort,
       &getSvrTime,
       &getSvrData,
       &getNumFreeSockets
};

/***********************************
     ROM constant strings
***********************************/
code const char httpHead[]           = "HTTP/1.1 200 OK\n" ;           // HTTP header
code const char httpDate[]           = "Date: Tue, 02 Jan 2013 08:12:31 GMT\n"; // Date: Tue, 15 Nov 1994 08:12:31 GMT
code const char httpServer[]         = "Server: alphaloewe 0.1.1\n";
code const char httpConnectionClose[]= "Connection: close\n";
code const char httpConnectionAlive[]= "Connection: Keep-Alive\n";
code const char httpConnectionKeep[] = "Keep-Alive: timeout=15, max=15\n";
code const char httpTransferEn[]     = "Transfer-Encodeing: chunked\n";
code const char httpPragmaDir[]      = "Pragma-directive: no-cache\n";
code const char httpCacheDir[]       = "Cache-directive: no-cache\n";
code const char httpCacheCon[]       = "Cache-control: no-store\n";
code const char httpPragma[]         = "Pragma: nocache\n";
code const char httpContentType[]    = "Content-Type: ";
code const char httpContentLength[]  = "Content-Length: ";
code const char httpMimeTypeHTML[]   = "text/html\n\n" ;                    // HTML MIME type
code const char httpMimeTypeCSS[]    = "text/css\n\n" ;                     // CSS MIME type
code const char httpMimeTypeJS[]     = "application/javascript\n\n";
code const char httpMimeTypeJPG[]    = "image/jpg\n\n";
code const char httpMimeTypePNG[]    = "image/png\n\n";
code const char httpNotSupported[]   = "HTTP/1.1 501 Not Implemented\n\n";
code const char errorPage[]          = "<!doctype html><html><head><meta charset=\"utf-8\"><title>Error</title></head><h1>Error</h1><body><p>The Webserver Cannot find its' media. Please insert the SD Card.</p></body><footer><a href=\"www.alphloewe.com\" title=\"Alphaloewe Home\">AlphaLoewe</a></footer></html>";


/*****************************
     Helper Function Proto
*****************************/
/*!
 * \brief Converts strings stored in program memory to ram
 *
 * \param const char *ctxt - flash based string
 */
static char* codeTxt2RamTxt(const char *ctxt);
/*!
 * \brief Starts file system
 *
 */
static void startFS(void);
/*!
 *  \brief Resets indexs of socket once closed
 *
 *  \param int socket - socket number to close
 *
 */
static void resetSocket(int socket);
/*!
 *  \brief Validates if the file exists
 *
 *  \param int file_t - the type of file to look for
 *  \param int index - the index of the websocket to tie to
 *  \param char *name - name of the file to look for
 *
 *  \return int - results of search
 *    \retval 1 = FOUND
 *    \retval 0 = NOT Found
 */
static int validateFile(int file_t, int index, char* name);

/***********************************
     Functions
***********************************/
// ISR that detects card insertion
void cardInsertISR() iv IVT_ADDR_PCINT2 ics ICS_AUTO {
     if(Button(&PINC, 5, 30, 0)) {
          sdCardInsert = 1;
     } else {
          sdCardInsert = 0;
          status = NOTREADY;
     }
}
// Validates file
static int validateFile(int file_t, int index, char* name){
     FAT32_ChangeDir("\\");
     FAT32_ChangeDir("WEB");
     switch (file_t){
          case FILE_CSS:
               FAT32_ChangeDir("STYLES");
          break;
          case FILE_JS:
               FAT32_ChangeDir("SCRIPTS");
          break;
          case FILE_PNG:
               FAT32_ChangeDir("IMAGES");
          break;
          case FILE_JPG:
               FAT32_ChangeDir("IMAGES");
          break;
     }
     if(FAT32_Exists(name) == 1){
          webSoc[index].filetype = file_t;
          webSoc[index].mark = 1;
          webSoc[index].position = 0;
          strcpy(&webSoc[index].fileName, name);
          FAT32_Size(webSoc[index].fileName, &webSoc[index].fileSize);
          webSoc[index].fileHandle = FAT32_Open(&webSoc[index].fileName, FILE_READ);
          if(webSoc[index].fileHandle == -1) logEvent(LOG_ERROR, "Filehandle Failed");
          // Charges the buffer first go around
          memset(&webSoc[index].buffer, 0, DATABUFFERSIZE);
          if(FAT32_Read(webSoc[index].fileHandle, &webSoc[index].buffer, DATABUFFERSIZE)==0){
               webSoc[index].bufPos = 0;
          } else {
               logEvent(LOG_ERROR, "Could not read from file");
               FAT32_ChangeDir("\\");
               return 0;
          }
          FAT32_ChangeDir("\\");
          return 1;
     } else {
          FAT32_ChangeDir("\\");
          return 0;
     }
}
// Resets socket indexes
static void resetSocket(int socNum){
       webSoc[socNum].position = 0;
       webSoc[socNum].mark = 0;
       webSoc[socNum].bufPos = 0;
       if(Net_Ethernet_28j60_disconnectTCP(webSoc[socNum].socket) == 0){
            logEvent(LOG_ERROR, "Failed to Close Socket\n");
       }
}
// Converts FLASH variables to RAM
static char *codeTxt2RamTxt(const char *ctxt){
     static char txt[20];
     int i = 0;

     for(i =0; txt[i] = ctxt[i]; i++);
     return &txt;
}
// Handles all TCP packets
void Net_Ethernet_28j60_UserTCP(SOCKET_28j60_Dsc *socket) {
     if(status == READY){                                           // If SD Card is Ready, else a error_webpage is displayed
          int i = 0;
          if(socket->destPort != portNumber){
               return;                                              // Listen only to web requests on programmed port
          } else {
               char tmpChr = 0;
               char request[REQBUFFERSIZE];
               char * const pRequest = &request;
               memset(pRequest, 0, REQBUFFERSIZE);                  // Set the buffer to 0, keeps from having to set '\0' at the end of the string

               for(i=0; i < REQBUFFERSIZE-1; i++){                  // Extract the request communication
                    tmpChr = Net_Ethernet_28j60_getByte();
                    if(tmpChr == '\n') break;
                    else *(pRequest+i) = tmpChr;
               }

               if(memcmp(pRequest, "GET ", 3) == 0){                // If request is a GET, break the request down into tokens
                    char commands[NUMCOMMAND][COMMANDSIZE];
                    char (*pCommands)[COMMANDSIZE] = &commands;
                    char strC[COMMANDSIZE];
                    char * const pStrC = &strC;
                    int n = 0;

                    strcpy(pStrC, strtok(pRequest, TOKENS));               // Get the first token
                    for(n=0; (pStrC[0] != '\0') && (n < NUMCOMMAND); n++){ // If there are tokens left in the string set them here
                         strcpy(pCommands[n], pStrC);                      // Copy the token to a set of commands
                         strcpy(pStrC, strtok(0, TOKENS));
                    }

                    if((strcmp(pCommands[0], "GET") == 0) && socket->state != 3){
                         return;
                    } else if (strcmp(pCommands[0], "GET") == 0){
                         char fileStr[COMMANDSIZE];
                         char * const pFileStr = &fileStr;

                         memset(pFileStr, 0, COMMANDSIZE);
                         for(i=0; i<n; i++){
                              if(strcmp(pCommands[i],"HTTP") == 0){
                                   if(i == 1) {
                                        pFileStr = "index.htm";
                                   } else {
                                        strcpy(pFileStr, pCommands[i-1]);
                                   break;
                                   }
                              }
                         }
                         for(i = 0; i < MAX_SOCKETS; i++){
                              if(webSoc[i].mark == 0 && webSoc[i].socket->open == 1){

                                   char *fEnd;
                                   char *res;
                                   char txt[7];

                                   fEnd = strchr(pFileStr, '.');        // Lets see what the extension is!

                                   if(strncmp(fEnd, ".htm", 3) == 0){   // Behavior is based on file extension
                                        if(validateFile(FILE_HTML, i ,pFileStr)){
                                             webSoc[i].socket = socket;
                                        } else {
                                             Net_Ethernet_28j60_disconnectTCP(socket);
                                             return;
                                        }
                                   } else if(strncmp(fEnd, ".css", 3) == 0){
                                        if(validateFile(FILE_CSS, i ,pFileStr)){
                                             webSoc[i].socket = socket;
                                        } else {
                                             Net_Ethernet_28j60_disconnectTCP(socket);
                                             return;
                                        }
                                   } else if(strncmp(fEnd, ".js", 3) == 0){
                                        if(validateFile(FILE_JS, i ,pFileStr)){
                                             webSoc[i].socket = socket;
                                        } else {
                                             Net_Ethernet_28j60_disconnectTCP(socket);
                                             return;
                                        }
                                   } else if(strncmp(fEnd, ".jpg", 3) == 0){
                                        if(validateFile(FILE_JPG, i ,pFileStr)){
                                             webSoc[i].socket = socket;
                                        } else {
                                             Net_Ethernet_28j60_disconnectTCP(socket);
                                             return;
                                        }
                                   } else if(strncmp(fEnd, ".png", 3) == 0){
                                        if(validateFile(FILE_PNG, i ,pFileStr)){
                                             webSoc[i].socket = socket;
                                        } else {
                                             Net_Ethernet_28j60_disconnectTCP(socket);
                                             return;
                                        }
                                   } else if(strncmp(fEnd, ".ico", 3) == 0){
                                        if(validateFile(FILE_ICO, i ,pFileStr)){
                                             webSoc[i].socket = socket;
                                        } else {
                                             Net_Ethernet_28j60_disconnectTCP(socket);
                                             return;
                                        }
                                   } else {
                                        return;
                                   }

                                   Net_Ethernet_28j60_putConstStringTCP(httpHead, webSoc[i].socket);
                                   Net_Ethernet_28j60_putConstStringTCP(httpServer, webSoc[i].socket);
                                   Net_Ethernet_28j60_putConstStringTCP(httpDate, webSoc[i].socket);
                                   Net_Ethernet_28j60_putConstStringTCP(httpConnectionClose, webSoc[i].socket);
                                   Net_Ethernet_28j60_putConstStringTCP(httpContentLength, webSoc[i].socket);
                                   LongToStr(webSoc[i].fileSize, txt);                         // Convert the size of the file to string
                                   res = Ltrim(txt);                                           // Trims the filesize to eliminate pre '\0's
                                   Net_Ethernet_28j60_putStringTCP(res, webSoc[i].socket);
                                   Net_Ethernet_28j60_putByteTCP('\n', webSoc[i].socket);
                                   switch(webSoc[i].fileType){
                                       case FILE_HTML:
                                            Net_Ethernet_28j60_putConstStringTCP(httpContentType, webSoc[i].socket);
                                            Net_Ethernet_28j60_putConstStringTCP(httpMimeTypeHTML, webSoc[i].socket);
                                       break;
                                       case FILE_CSS:
                                            Net_Ethernet_28j60_putConstStringTCP(httpContentType, webSoc[i].socket);
                                            Net_Ethernet_28j60_putConstStringTCP(httpMimeTypeCSS, webSoc[i].socket);
                                       break;
                                       case FILE_JS:
                                            Net_Ethernet_28j60_putConstStringTCP(httpContentType, webSoc[i].socket);
                                            Net_Ethernet_28j60_putConstStringTCP(httpMimeTypeJS, webSoc[i].socket);
                                       break;
                                       case FILE_PNG:
                                            Net_Ethernet_28j60_putConstStringTCP(httpContentType, webSoc[i].socket);
                                            Net_Ethernet_28j60_putConstStringTCP(httpMimeTypePNG, webSoc[i].socket);
                                       break;
                                       case FILE_JPG:
                                            Net_Ethernet_28j60_putConstStringTCP(httpContentType, webSoc[i].socket);
                                            Net_Ethernet_28j60_putConstStringTCP(httpMimeTypeJPG, webSoc[i].socket);
                                       break;
                                       case FILE_ICO:
                                            Net_Ethernet_28j60_putConstStringTCP(httpContentType, webSoc[i].socket);
                                            Net_Ethernet_28j60_putConstStringTCP(httpMimeTypePNG, webSoc[i].socket);
                                       break;
                                   }
                                   break;
                              }
                         }
                    }
               }
          }

          /* Handle HTML request only if socket is correct and mark is set */
          for(i = 0; i < MAX_SOCKETS; i++){
               if((webSoc[i].mark == 1) && (webSoc[i].socket == socket)){
                    while(webSoc[i].position < webSoc[i].fileSize){
                         if((webSoc[i].bufPos == DATABUFFERSIZE) && (FAT32_Eof(webSoc[i].fileHandle) != 1)) {
                              FAT32_Read(webSoc[i].fileHandle, &webSoc[i].buffer, DATABUFFERSIZE);
                              webSoc[i].bufPos = 0;
                         }
                         if(FAT32_Eof(webSoc[i].fileHandle) == 1) {
                              FAT32_Close(webSoc[i].fileHandle);
                         }
                         switch(Net_Ethernet_28j60_putByteTCP(webSoc[i].buffer[webSoc[i].bufPos], webSoc[i].socket)){
                              case 1:
                                   webSoc[i].bufPos++; webSoc[i].position++;
                              break;
                              case 0:
                                   return;
                              break;
                         }
                    }
                    if((Net_Ethernet_28j60_bufferEmptyTCP(webSoc[i].socket) == 1) && webSoc[i].position >= webSoc[i].fileSize){
                         resetSocket(i);
                    }
               }
          }

     } else {
          // No SD Card inserted
          char *p;
          if(webSoc[FILE_HTML].position == 0){
               webSoc[FILE_HTML].mark = 1;
               webSoc[FILE_HTML].socket = socket;
               webSoc[FILE_HTML].bufPos = 0;
               webSoc[FILE_HTML].fileSize = 264;
               webSoc[FILE_HTML].position += Net_Ethernet_28j60_putConstStringTCP(httpHead, webSoc[FILE_HTML].socket);
               webSoc[FILE_HTML].position += Net_Ethernet_28j60_putConstStringTCP(httpServer, webSoc[FILE_HTML].socket);
               webSoc[FILE_HTML].position += Net_Ethernet_28j60_putConstStringTCP(httpDate, webSoc[FILE_HTML].socket);
               webSoc[FILE_HTML].position += Net_Ethernet_28j60_putConstStringTCP(httpConnectionClose, webSoc[FILE_HTML].socket);
               webSoc[FILE_HTML].position += Net_Ethernet_28j60_putConstStringTCP(httpContentLength, webSoc[FILE_HTML].socket);
               webSoc[FILE_HTML].position += Net_Ethernet_28j60_putStringTCP("264", webSoc[FILE_HTML].socket);
               webSoc[FILE_HTML].position += Net_Ethernet_28j60_putByteTCP('\n', webSoc[FILE_HTML].socket);
               webSoc[FILE_HTML].position += Net_Ethernet_28j60_putConstStringTCP(httpContentType, webSoc[FILE_HTML].socket);
               webSoc[FILE_HTML].position += Net_Ethernet_28j60_putConstStringTCP(httpMimeTypeHTML, webSoc[FILE_HTML].socket);
          }

          while(webSoc[FILE_HTML].bufPos < webSoc[FILE_HTML].fileSize && socket == webSoc[FILE_HTML].socket) {
               *p = errorPage[webSoc[FILE_HTML].bufPos];
               if(Net_Ethernet_28j60_putByteTCP(*p, webSoc[FILE_HTML].socket) == 1){
                              webSoc[FILE_HTML].position++;
                              webSoc[FILE_HTML].bufPos++;
               } else {
                    break;
               }
          }

          if((Net_Ethernet_28j60_bufferEmptyTCP(webSoc[FILE_HTML].socket) == 1) && (webSoc[FILE_HTML].bufPos >= webSoc[FILE_HTML].fileSize) && (socket == webSoc[FILE_HTML].socket)){
               Net_Ethernet_28j60_disconnectTCP(webSoc[FILE_HTML].socket);
               resetSocket(FILE_HTML);
          }
     }
}
// Handles all UDP packets
unsigned int Net_Ethernet_28j60_UserUDP(UDP_28j60_Dsc *udpDsc) {
     // Port 123 is a SNTP request
     if(udpDsc->destPort == SNTP_PORT){
          struct SNTP_Response tmResp;                //!< Holds packet framework for SNTP servers
          unsigned long tts = 0;                      //!< Holds the time returned from the remote SNTP
          int i = 0;

          tmResp.serverFlags = Net_Ethernet_28j60_getByte() ;
          tmResp.serverStratum = Net_Ethernet_28j60_getByte() ;
          Net_Ethernet_28j60_getByte() ;              // skip poll
          tmResp.serverPrecision = Net_Ethernet_28j60_getByte() ;

          for(i = 0 ; i < 36 ; i++){
                Net_Ethernet_28j60_getByte() ;        // skip all unused fileds
          }

          // store transmit timestamp
          Highest(tts) = Net_Ethernet_28j60_getByte() ;
          Higher(tts) = Net_Ethernet_28j60_getByte() ;
          Hi(tts) = Net_Ethernet_28j60_getByte() ;
          Lo(tts) = Net_Ethernet_28j60_getByte() ;

          // convert sntp timestamp to unix epoch
          tmResp.epoch = tts - 2208988800;            // The large number is the difference between 1900 and 1970
          //TimeRTC.TimeSync(&tmResp.epoch);
          logEvent(LOG_NORMAL, "SNTP Time Sync");
          return (udpDsc->dataLength);
     }
     // TODO: This is where a file download will take place
     return 0;
}
// Sets the internal IP address
void setIP(ipAddress_t *address){
     myIpAddr[0] = address->ip1;
     myIpAddr[1] = address->ip2;
     myIpAddr[2] = address->ip3;
     myIpAddr[3] = address->ip4;
}
// Sets the SUBNET mask ip address
void setSubIP(ipAddress_t *address){
     ipMask[0] = address->ip1;
     ipMask[1] = address->ip2;
     ipMask[2] = address->ip3;
     ipMask[3] = address->ip4;
}
// Sets the Gateway ip address
void setGW(ipAddress_t *address){
     gwIpAddr[0] = address->ip1;
     gwIpAddr[1] = address->ip2;
     gwIpAddr[2] = address->ip3;
     gwIpAddr[3] = address->ip4;
}
// Sets the DNS ip address
void setDNS(ipAddress_t *address){
     dnsIpAddr[0] = address->ip1;
     dnsIpAddr[1] = address->ip2;
     dnsIpAddr[2] = address->ip3;
     dnsIpAddr[3] = address->ip4;
}
// Set the port that the server is listening on
void setPort(unsigned short portNum){
     portNumber = portNum;
}
// Sets the DHCP mode
void setDHCP(int mode){
     dhcp_mode = mode;
}
// Logger that writes ERROR messages to file
void logEvent(int log, char* text){
    UART_Write_Text(text);
    UART_Write('\n');
    /*__HANDLE fileHandleLog;

     if(status==READY){
          __TIME tm;
          static char tempTxt[100];
          char tmTxt[9];
          char * const pTempTxt = &tempTxt;
          memset(pTempTxt, 0 , 100);
          // Convert RTC BCD format
          tm.Year = 2000 + Bcd2Dec(TimeRTC.year);
          tm.Month = Bcd2Dec(TimeRTC.month);
          tm.Day = Bcd2Dec(TimeRTC.day);
          tm.Hour = Bcd2Dec(TimeRTC.hour);
          tm.Minute = Bcd2Dec(TimeRTC.mins);
          tm.Second = Bcd2Dec(TimeRTC.sec);
          FAT32_SetTime(&tm);
          // Get the formatted string date from TimeRTC and concat the string
          strcpy(tmTxt, TimeRTC.GetDate());
          pTempTxt = strcat(pTempTxt, tmTxt);
          pTempTxt = strcat(pTempTxt, " : ");
          strcpy(tmTxt, TimeRTC.GetTime(mode12Hour));
          pTempTxt = strcat(pTempTxt, tmTxt);
          pTempTxt = strcat(pTempTxt, " - ");
          pTempTxt = strcat(pTempTxt, text);
          pTempTxt = strcat(pTempTxt, "\r\n");
          switch (log){
               case LOG_NORMAL:
                    fileHandleLog = FAT32_Open("SYSTEM.TXT", FILE_APPEND);
               break;
               case LOG_ERROR:
                    fileHandleLog = FAT32_Open("ERROR.TXT", FILE_APPEND);
               break;
               case LOG_UNKNOWN:
                    fileHandleLog = FAT32_Open("UNKNOWN.TXT", FILE_APPEND);
               break;
          }
          FAT32_Write(fileHandleLog, pTempTxt, strlen(pTempTxt));
          FAT32_Close(fileHandleLog);
     }*/
}
// Retrieves SNTP timestamp from input
void getSvrTime(unsigned char *ntpSrv){
     unsigned char sntpPkt[50];
     unsigned char remoteIpAddr[4];

     memcpy(remoteIpAddr, Net_Ethernet_28j60_dnsResolve(ntpSrv, 4), 4);

     memset(sntpPkt, 0, 48);        // clear sntp packet
     // FLAGS : byte 0
     sntpPkt[0] = 0b00011001;       // LI = 0 ; VN = 3 ; MODE = 1
     // STRATUM : byte 1 = 0
     sntpPkt[1] = 0x00;
     // POLL : byte 2
     sntpPkt[2] = 0x0a;             // 1024 sec (arbitrary value)
     // PRECISION : byte 3
     sntpPkt[3] = 0xfa;             // 0.015625 sec (arbitrary value)
     // DELAY : bytes 4 to 7 = 0.2656 sec (arbitrary value)
     sntpPkt[6] = 0x44 ;
     // DISPERSION : bytes 8 to 11 = 16 sec (arbitrary value)
     sntpPkt[9] = 0x10 ;
     // REFERENCE ID : bytes 12 to 15 = 0 (unspecified)

     // REFERENCE TIMESTAMP : bytes 16 to 23 (unspecified)

     // ORIGINATE TIMESTAMP : bytes 24 to 31 (unspecified)

     // RECEIVE TIMESTAMP : bytes 32 to 39 (unspecified)

     // TRANSMIT TIMESTAMP : bytes 40 to 47 (unspecified)

     Net_Ethernet_28j60_sendUDP(remoteIpAddr, SNTP_PORT, SNTP_PORT, sntpPkt, 48);

}
// Returns data requested from server
char* getSvrData(unsigned char *request){
     /* Example request:
     http://tidesandcurrents.noaa.gov/noaatidepredictions/
     viewMonthlyPredictions.jsp?bmon=12&bday=30&byear=2012&timelength
     =monthly&timeZone=2&dataUnits=1&datum=MLLW&timeUnits=2&interval=
     highlow&Threshold=greaterthanequal&thresholdvalue=&format=Submit&Stationid=9432780 */
     //TODO: Parse to / find address
     //TODO: resolveDNS
     SOCKET_28j60_Dsc *socket_global;
     int i;

     i = Net_Ethernet_28j60_connectTCP(request, 80, 10000, &socket_global);
     Net_Ethernet_28j60_startSendTCP(socket_global);
}
// Read packets as they come in
void doPacket(){
     Net_Ethernet_28j60_doPacket();
     if(sdCardInsert == 1) {
          startFS();
     }
     if(Net_Ethernet_28j60_doDHCPLeaseTime() && dhcp_mode == DHCP_ON){
          Net_Ethernet_28j60_renewDHCP(5); // it's time to renew the IP address lease, with 5 secs for a reply
     }
}
// Gets the number of free sockets
int getNumFreeSockets(){
    int i = 0, count = 0;

    for(i=0; i< MAX_SOCKETS; i++){
         if(webSoc[i].socket->open == 1) count++;
         if(webSoc[i].mark = 0 && webSoc[i].socket->open == 0) Net_Ethernet_28j60_disconnectTCP(webSoc[i].socket);
    }
    return count;
}
// Start the File System
static void startFS(){
     if(Mmc_Insert_Detect == 0){
          switch (FAT32_Init()){
               case 0:
                    Delay_ms(50);
                    status = READY;
                    sdCardInsert = 0;
                    UART_Write_Text("FS Ready\n");
               break;
               default:
                    status = NOTREADY;
               break;
          }
     } else {
          status = NOTREADY;
          UART_Write_Text("No card detected\n");
     }
}
// Initializes the SD Card
void SDCardInit(){
     status = NOTREADY;
     PCICR |= (1<<PCIE2);                                 // Enable external pin change interrupt
     PCMSK2 |= (1<<PCINT21);
     Mmc_Insert_Direction = 0;
     startFS();
}
// Initializes the webserver
void webserverInit(void) {
     int i = 0;
     Net_Ethernet_28j60_stackInitTCP();                                    // Initializes the TCP stack
     Net_Ethernet_28j60_Init(myMacAddr, myIpAddr, Net_Ethernet_28j60_FULLDUPLEX);
     // DHCP Mode
     if(dhcp_mode == DHCP_ON){
          if(Net_Ethernet_28j60_initDHCP(5) == 1){
               memcpy(myIpAddr, Net_Ethernet_28j60_getIpAddress(), 4);     // fetch IP address
               memcpy(gwIpAddr, Net_Ethernet_28j60_getGwIpAddress(), 4);   // fetch gateway IP address
               memcpy(dnsIpAddr, Net_Ethernet_28j60_getDnsIpAddress(), 4); // fetch DNS server address
               memcpy(ipMask, Net_Ethernet_28j60_getIpMask(), 4);          // fetch IP subnet mask
          } else {
               logEvent(LOG_ERROR, "Failed to connect to DHCP server");
          }
     }
     Net_Ethernet_28j60_confNetwork(myIpAddr, gwIpAddr, dnsIpAddr);
     Delay_ms(1000);                                                       //Allow time for the ethernet chip to stabilize

     for(i = 0; i < MAX_SOCKETS; i++){
          webSoc[i].position = 0;
          webSoc[i].mark = 0;
          memset(webSoc[i].fileName, 0, COMMANDSIZE);
     }
}