/*!
 * \file
 *
 * \brief AVR Webserver
 *
 * Copyright:
 *   (c) Richard, 12/23/2012
 * Revision History:
 *   .01 - Working
 *   .02 - Added time request function
 *   .03 - Added number of free sockets function
 *   .04 - Changed method of closing sockets, added DHCP configuration as well as renew functions
 *
 * Status:
 *   51% completed.
 *
 * Test configuration:
 *   MCU:             ATMega1284p
 *   Dev.Board:       AVRPLC16
 *   Oscillator:      8Mhz
 *   Ext. Modules:    x
 *   SW:              MikroC 5.8
 * NOTES:
 *
 */

 /**
 * \defgroup web_group Web Server
 *
 *  @{
 */


#ifndef WEBSERVER_H
#define WEBSERVER_H

#include "__NetEthEnc28j60.h"


#define DEBUG
// Webserver
#define MAX_SOCKETS 5
#define DATABUFFERSIZE 512  /*!< Defines the buffer size used by each webSocket_desc_t to read from the SD Card */
#define REQBUFFERSIZE  100  /*!< Defines the request buffer size used when examining an incoming packet */
#define NUMCOMMAND 7        /*!< Number of commands accepted GET | / | HTTP | 1.1 | etc. */
#define COMMANDSIZE 13      /*!< Since our filesystem can only process 8.3 filenames, commands are limited to 12 chars */
#define TOKENS " /"         /*!< Tokens used to parse incoming packet request */
#define SNTP_PORT 123       /*!< Port that is used for SNTP*/

// SD Card Status
#define READY 1
#define NOTREADY 0


/********************************
     Globals
********************************/
/*!
 * \brief Hold log types
 *
 */
enum {
     LOG_NORMAL = 0,
     LOG_ERROR,
     LOG_UNKNOWN
};
/*!
 * \brief File handling types
 *
 */
enum {
     FILE_HTML = 0,
     FILE_CSS,
     FILE_JS,
     FILE_PNG,
     FILE_ICO,
     FILE_JPG,
     FILE_LOG
};
/*!
 * \brief Modes of operation for DHCP
 *
 */
enum {
     DHCP_OFF = 0,
     DHCP_ON
};
/*!
 * \brief Holds IP addresses
 *
 */
struct ipAddress_desc{
     unsigned char ip1;
     unsigned char ip2;
     unsigned char ip3;
     unsigned char ip4;
};

typedef struct ipAddress_desc ipAddress_t;

/*!
 * \brief Websocket struct that holds all function pointers
 *
 */
struct webServer_desc {
     const void(*Init)(void);
     const void(*FSInit)(void);
     const void(*DoPacket)(void);
     const void(*LogEvent)(int log, char* text);
     const void(*SetIP)(ipAddress_t *address);
     const void(*SetSubIP)(ipAddress_t *address);
     const void(*SetGW)(ipAddress_t *address);
     const void(*SetDNS)(ipAddress_t *address);
     const void(*SetDHCP)(int mode);
     const void(*SetPort)(unsigned short portNum);
     const void(*GetTime)(unsigned char *ntpSvr);
     const char*(*GetData)(unsigned char *request);
     const int(*GetFreeSockets)(void);
};

/*****************************
     Prototypes
*****************************/
/*!
 * \brief Initializes webserver
 *
 */
void webserverInit(void);
/*!
 * \brief Initializes SD Card
 *
 */
void SDCardInit(void);
/*!
 * \brief Do packet called as much as possible
 *
 */
void doPacket(void);
/*!
 * \brief Logs event to SD card
 * \param int log is an enum type of log determines the log entry
 * \param char * text accepts string to be written to the log
 *
 */
void logEvent(int log, char* text);
/*!
 * \brief Sets the IP Address for the server
 * \param ipAddress_t *address for IP
 */
void setIP(ipAddress_t *address);
/*!
 * \brief Sets the SUBnet for the server
 * \param ipAddress_t *address for IP
 */
void setSubIP(ipAddress_t *address);
/*!
 * \brief Sets the Gateway for the server
 * \param ipAddress_t *address for IP
 */
void setGW(ipAddress_t *address);
/*!
 * \brief Sets the DNS for the server
 * \param ipAddress_t *address for IP
 */
void setDNS(ipAddress_t *address);
/*!
 * \brief Sets the mode for DHCP operation
 * \param int mode Accepts a enum type of operations
 */
void setDHCP(int mode);
/*!
 * \brief Sets the communication port for the server
 * \unsigned short portNum is the port number that the server will communicate over
 */
void setPort(unsigned short portNum);
/*!
 *  \brief Retrieves NTP timestamp from remote server
 *
 *  \param unsigned char *ntpSrv - web address of SNTP server e.g. "www.sntp.gov"
 *
 */
extern void getSvrTime(unsigned char *ntpSrv);
/*!
 *  \brief <Basic Description>
 *
 *  \pre <Preconditions that need to be met before called>
 *
 *  \param <Parameters Accepted>
 *
 *  \return <Returns>
 *    \retval <Values that might be returned>
 *
 *  NOTES:
 *   <notes>
 */
extern char* getSvrData(unsigned char *request);
/*!
 *  \brief Returns number of free sockets on server
 *
 *  \return int - number of free sockets
 *
 */
extern int getNumFreeSockets(void);

/*******************************************
 * Globals
 ******************************************/
extern struct webServer_desc Webserver;    //!< Holds all pointer functions for webserver

#endif